"""
Test Events module.

This module contains tests for Event objects.
"""
import datetime

import pytest

from esak import exceptions


def test_known_event(talker):
    se = talker.event(336)
    assert se.title == "Secret Empire"
    assert se.start == datetime.date(2017, 4, 19)
    assert se.end == datetime.date(2017, 8, 9)
    assert se.resource_uri == "http://gateway.marvel.com/v1/public/events/336"
    assert se.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/6/f0/58e692d6f351b.jpg"

    assert len(se.stories) == 20
    assert se.stories[0].id == 109335
    assert se.stories[0].name == "cover from Deadpool (2012) #32"
    assert se.stories[0].type == "cover"

    assert len(se.characters) == 20
    assert se.characters[0].id == 1011227
    assert se.characters[0].name == "Amadeus Cho"
    assert (
        se.characters[0].resource_uri
        == "http://gateway.marvel.com/v1/public/characters/1011227"
    )

    assert len(se.creators) == 20
    assert se.creators[0].id == 11482
    assert se.creators[0].name == "Jesus Aburtov"
    assert se.creators[0].role == "colorist (cover)"

    assert se.next.id == 322
    assert se.next.name == "Avengers NOW!"
    assert se.previous.id == 332
    assert se.previous.name == "Dead No More: The Clone Conspiracy"

    assert len(se.comics) == 20
    assert se.comics[0].id == 63223
    assert se.comics[0].name == "All-New Guardians of the Galaxy Annual (2017) #1"
    assert se.comics[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/63223"

    assert len(se.series) == 20
    assert se.series[11].id == 25356
    assert se.series[11].name == "Secret Empire Omega (2017)"
    assert se.series[11].resource_uri == "http://gateway.marvel.com/v1/public/series/25356"


def test_bad_event(talker):
    with pytest.raises(exceptions.ApiError):
        talker.event(-1)


def test_events_list(talker):
    events_lst = talker.events_list(
        {
            "orderBy": "modified",
        }
    )

    stories_iter = iter(events_lst)
    assert (next(stories_iter).id) == 296
    assert (next(stories_iter).id) == 302
    assert (next(stories_iter).id) == 233
    assert len(events_lst) == 20



"""
Test Creator module.
This module contains tests for Creator objects.
"""

from datetime import date
from decimal import Decimal

import pytest

from esak import exceptions


def test_known_creator(talker):
    jason = talker.creator(11463)
    assert jason.first_name == "Jason"
    assert jason.last_name == "Aaron"
    assert jason.id == 11463
    assert jason.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/7/10/5cd9c7870670e.jpg"

    assert 16450 in [s.id for s in jason.series]
    assert len(jason.series[:5]) == 5
    assert len(jason.series) == len([x for x in jason.series if x.id > 3])

    assert len(jason.stories) == 20
    assert jason.stories[0].id == 32907
    assert jason.stories[0].name == "Man In The Pit 1 of 1"
    assert jason.stories[0].type == "interiorStory"
    assert jason.stories[0].resource_uri == "http://gateway.marvel.com/v1/public/stories/32907"

    assert len(jason.comics) == 20
    assert jason.comics[0].id == 45762
    assert jason.comics[0].name == "A+X Vol. 1: = Awesome (Trade Paperback)"
    assert jason.comics[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/45762"

    assert len(jason.series) == 20
    assert jason.series[11].id == 24229
    assert jason.series[11].name == "Avengers (2018 - Present)"
    assert jason.series[11].resource_uri == "http://gateway.marvel.com/v1/public/series/24229"


def test_bad_creator(talker):
    with pytest.raises(exceptions.ApiError):
        talker.creator(-1)


def test_pulls_verbose(talker):
    creators = talker.creators_list(
        {
            "orderBy": "modified",
        }
    )
    c_iter = iter(creators)
    assert next(c_iter).full_name == "Sean Cooke"
    assert next(c_iter).full_name, "Mark Shultz"
    assert next(c_iter).full_name, "Miles Lane"
    assert len(creators) > 0


def test_creator_comics(talker):
    jason = talker.creator_comics(11463)
    assert len(jason.comics) == 20
    val5 = jason.comics[7]
    assert val5.id == 93341
    assert val5.series.id == 31903
    assert val5.series.name == "The Mighty Valkyries (2021)"
    assert val5.series.resource_uri == "http://gateway.marvel.com/v1/public/series/31903"
    assert val5.title == "The Mighty Valkyries (2021) #5"
    assert val5.issue_number == 5
    assert val5.page_count == 32
    assert val5.upc == "75960620098600511"
    assert val5.diamond_code == "JUN210717"
    assert val5.format == "Comic"
    assert len(val5.characters) == 2
    assert len(val5.creators) == 5
    assert len(val5.events) == 0
    assert len(val5.stories) == 2
    assert val5.prices.digital is None
    assert val5.prices.print == Decimal("3.99")
    assert val5.dates.on_sale == date(2021, 9, 15)
    assert val5.dates.foc == date(2021, 8, 23)


def test_creator_events(talker):
    jason = talker.creator_events(11463)
    assert len(jason.events) == 10
    s = jason.events[5]
    assert s.id == 309
    assert s.title == "Shattered Heroes"
    assert s.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/2/a0/511e8000770cd.jpg"
    assert s.resource_uri == "http://gateway.marvel.com/v1/public/events/309"
    assert len(s.characters) == 20
    assert len(s.comics) == 20
    assert len(s.creators) == 20
    assert len(s.series) == 10
    assert len(s.stories) == 20
    assert s.start == date(2011, 10, 19)
    assert s.end == date(2012, 4, 22)


def test_creator_series(talker):
    jason = talker.creator_series(11463)
    assert len(jason.series) == 20
    ax = jason.series[0]
    assert ax.id == 16450
    assert ax.start_year == 2012
    assert ax.end_year == 2014
    assert ax.title == "A+X (2012 - 2014)"
    assert ax.next is None
    assert ax.previous is None


def test_creator_stories(talker):
    jason = talker.creator_stories(11463)
    assert len(jason.stories) == 20
    man = jason.stories[0]
    assert man.id == 32907
    assert man.title == "Man In The Pit 1 of 1"
    assert man.type == "story"
    assert man.resource_uri == "http://gateway.marvel.com/v1/public/stories/32907"
    assert len(man.characters) == 2
    assert len(man.comics) == 1
    assert len(man.creators) == 4
    assert man.original_issue.id == 16112
    assert man.original_issue.name == "Wolverine (2003) #56"



"""
Conftest module.
This module contains pytest fixtures.
"""

import os

import pytest

from esak import api, sqlite_cache


@pytest.fixture(scope="session")
def dummy_pubkey():
    """public key fixture."""
    return os.getenv("PUBLIC_KEY", "pub")


@pytest.fixture(scope="session")
def dummy_privkey():
    """private key fixture."""
    return os.getenv("PRIVATE_KEY", "priv")


@pytest.fixture(scope="session")
def talker(dummy_pubkey, dummy_privkey):
    """esak api fixture."""
    return api(
        public_key=dummy_pubkey,
        private_key=dummy_privkey,
        cache=sqlite_cache.SqliteCache("tests/testing_mock.sqlite"),
    )



"""
Test Init module.
This module contains tests for project init.
"""
import pytest

from esak import api, exceptions, session


def test_api():
    with pytest.raises(exceptions.AuthenticationError):
        api()

    with pytest.raises(exceptions.AuthenticationError):
        api(private_key="Something")

    with pytest.raises(exceptions.AuthenticationError):
        api(public_key="Something")

    m = None
    try:
        m = api(public_key="Something", private_key="Else")
    except Exception as exc:
        print(f"mokkari.api() raised {exc} unexpectedly!")

    assert m.__class__.__name__, session.Session.__name__



"""
Comic module.

This module provides the following classes:

- Comic
- ComicSchema
- ComicsList
"""
import itertools

from marshmallow import INCLUDE, Schema, fields, post_load, pre_load
from marshmallow.exceptions import ValidationError

from esak import dates, exceptions, generic_summary, prices, series, urls


class Comic:
    """
    The Comic object contains information for a comic.

    :param `**kwargs`: The keyword arguments is used for setting comic data from Marvel.
    """

    def __init__(self, **kwargs):
        """Intialize a new comic."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class ComicSchema(Schema):
    """Schema for the Comic API."""

    id = fields.Int()
    digitalId = fields.Int(attribute="digital_id")
    title = fields.Str()
    issueNumber = fields.Int(attribute="issue_number")
    variantDescription = fields.Str(attribute="variant_description")
    description = fields.Str(allow_none=True)
    modified = fields.DateTime()
    isbn = fields.Str()
    upc = fields.Str()
    diamondCode = fields.Str(attribute="diamond_code")
    ean = fields.Str()
    issn = fields.Str()
    format = fields.Str()
    pageCount = fields.Int(attribute="page_count")
    # textObjects
    resourceURI = fields.Url(attribute="resource_uri")
    urls = fields.Nested(urls.UrlsSchema)
    series = fields.Nested(series.SeriesSchema)
    variants = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    collections = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    collectedIssues = fields.Nested(
        generic_summary.GenericSummarySchema, attribute="collected_issues", many=True
    )
    dates = fields.Nested(dates.DatesSchema)
    prices = fields.Nested(prices.PriceSchemas, allow_none=True)
    # thumbnail
    images = fields.List(fields.Url)
    creators = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    characters = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    stories = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    events = fields.Nested(generic_summary.GenericSummarySchema, many=True)

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """
        Clean the data from Marvel.

        :param data: Data from Marvel response.

        :returns: Marvel Response
        :rtype: dict
        """
        if data.get("code", 200) != 200:
            raise exceptions.ApiError(data.get("status"))

        if "status" in data:
            data = data["data"]["results"][0]

        # Marvel comic 1768, and maybe others, returns a modified of
        # "-0001-11-30T00:00:00-0500". The best way to handle this is
        # probably just to ignore it, since I don't know how to fix it.
        if data.get("modified", " ")[0] == "-":
            del data["modified"]

        if "stories" in data:
            data["stories"] = data["stories"]["items"]

        if "events" in data:
            data["events"] = data["events"]["items"]

        if "creators" in data:
            data["creators"] = data["creators"]["items"]

        if "characters" in data:
            data["characters"] = data["characters"]["items"]

        if "images" in data:
            data["images"] = [f"{img['path']}.{img['extension']}" for img in data["images"]]

        if "isbn" in data:
            data["isbn"] = str(data["isbn"])

        if "diamondCode" in data:
            data["diamondCode"] = str(data["diamondCode"])

        return data

    @post_load
    def make(self, data, **kwargs):
        """
        Make the Comic object.

        :param data: Data from Marvel response.

        :returns: :class:`Comic` object
        :rtype: Comic
        """
        return Comic(**data)


class ComicsList:
    """The ComicsList object contains a list of `Comic` objects."""

    def __init__(self, response):
        """Initialize a new ComicList."""
        self.comics = []

        for comic_dict in response["data"]["results"]:
            try:
                result = ComicSchema().load(comic_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.comics.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.comics)

    def __len__(self):
        """Return the length of the object."""
        return len(self.comics)

    def __getitem__(self, index):
        """Return the object of a at index."""
        try:
            return next(itertools.islice(self.comics, index, index + 1))
        except TypeError:
            return list(itertools.islice(self.comics, index.start, index.stop, index.step))



"""
Dates module.

This module provides the following classes:

- Dates
- DatesSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load, pre_load


class Dates:
    """
    The Dates object contains common dates.

    :param `on_sale`: The date the comic went on sale.
    :param `foc`: The final order cutoff date.
    :param `unlimited`: The date it was release on Marvel Unlimited.
    :param `**kwargs`: The keyword arguments used for setting any other date from Marvel.
    """

    def __init__(self, on_sale=None, foc=None, unlimited=None, **kwargs):
        """Intialize a new date."""
        self.on_sale = on_sale
        self.foc = foc
        self.unlimited = unlimited
        self.unknown = kwargs


class DatesSchema(Schema):
    """Schema for the Dates."""

    onsaleDate = fields.Date(attribute="on_sale")
    focDate = fields.Date(attribute="foc")
    unlimitedDate = fields.Date(attribute="unlimited")

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        dateformat = "%Y-%m-%dT%H:%M:%S%z"

    @pre_load
    def process_input(self, data, **kwargs):
        """Clean up response data."""
        # Marvel comic 4373, and maybe others, returns a focDate of
        # "-0001-11-30T00:00:00-0500". The best way to handle this is
        # probably just to ignore it, since I don't know how to fix it.
        return {d["type"]: d["date"] for d in data if d["date"][0] != "-"}

    @post_load
    def make(self, data, **kwargs):
        """
        Make the dates object.

        :param data: Data from Marvel response.

        :returns: :class:`Dates` object
        :rtype: Dates
        """
        return Dates(**data)



"""
Generic Summary module.

This module provides the following classes:

- GenericSummary
- GenericSummarySchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load, pre_load


class GenericSummary:
    """
    The GenericSummary object contains basic information.

    :param `**kwargs`: The keyword arguments used for getting data from Marvel.
    """

    def __init__(self, id=None, name=None, resource_uri=None, type=None, role=None, **kwargs):
        """Intialize a new GenericSummary."""
        self.id = id
        self.name = name
        self.resource_uri = resource_uri
        self.type = type
        self.role = role
        self.unknown = kwargs


class GenericSummarySchema(Schema):
    """Schema for the GenericSummary."""

    id = fields.Int()
    name = fields.Str()
    resourceURI = fields.Str(attribute="resource_uri")
    type = fields.Str()
    role = fields.Str()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """Extract the summary id."""
        data["id"] = data["resourceURI"].split("/")[-1]
        return data

    @post_load
    def make(self, data, **kwargs):
        """
        Make the GenericSummary object.

        :param data: Data from Marvel response.

        :returns: :class:`GenericSummary` object
        :rtype: GenericSummary
        """
        return GenericSummary(**data)



"""
Test Comic module.
This module contains tests for Comic objects.
"""
from datetime import date
from decimal import Decimal


def test_pulls_verbose(talker):
    week = talker.comics_list(
        {
            "format": "comic",
            "formatType": "comic",
            "noVariants": True,
            "dateDescriptor": "thisWeek",
        }
    )
    c_iter = iter(week)
    assert next(c_iter).id == 89738
    assert next(c_iter).id == 93316
    assert next(c_iter).id == 93320
    assert len(week) > 0


def test_pulls_simple(talker):
    week = talker.comics_list({"dateDescriptor": "thisWeek"})
    assert len(week.comics) > 0


def test_pulls_simpler(talker):
    week = talker.comics_list()
    assert len(week.comics) > 0


def test_known_comic(talker):
    af15 = talker.comic(16926)
    assert af15.title == "Amazing Fantasy (1962) #15"
    assert af15.issue_number == 15
    assert af15.description is None
    assert af15.format == "Comic"
    assert af15.id == 16926
    assert "Spider-Man (Peter Parker)" in [c.name for c in af15.characters]
    assert "Foo" not in [c.name for c in af15.characters]
    assert "Steve Ditko" in [s.name for s in af15.creators]
    assert "Abe Lincoln" not in [s.name for s in af15.creators]
    assert af15.prices.print == Decimal("0.00")
    assert af15.prices.digital is None
    assert len(af15.stories) == 5
    assert af15.stories[0].id == 35504
    assert af15.stories[0].type == "cover"
    assert af15.stories[0].name == "Cover: Amazing Fantasy (1962) #15"
    assert af15.stories[1].id == 35505
    assert af15.stories[1].type == "interiorStory"
    assert af15.stories[1].name == "Spider-Man!"
    assert af15.stories[2].id == 35506
    assert af15.stories[2].type == "interiorStory"
    assert af15.stories[2].name == "The Bell-Ringer"
    assert af15.dates.on_sale == date(1962, 8, 1)
    assert af15.dates.foc is None
    assert af15.dates.unlimited == date(2008, 5, 13)
    assert len(af15.collections) == 1
    assert af15.collections[0].id == 16214
    assert af15.collections[0].name == "AMAZING FANTASY OMNIBUS HC (Hardcover)"
    assert (
        af15.collections[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/16214"
    )
    assert len(af15.collected_issues) < 1
    assert len(af15.variants) < 1


def test_invalid_isbn(talker):
    """Sometimes Marvel API sends number for isbn"""
    murpg = talker.comic(1143)
    assert murpg.isbn == "785110283"
    assert murpg.prices.print == Decimal("9.99")
    assert murpg.prices.digital is None


def test_invalid_diamond_code(talker):
    """Sometimes Marvel API sends number for diamond code"""
    hulk = talker.comic(27399)
    assert hulk.diamond_code == "0"


def test_upc_code(talker):
    cable = talker.comic(95781)
    assert cable.upc == "759606201991000111"
    assert cable.dates.on_sale == date(2021, 8, 25)
    assert cable.dates.foc == date(2021, 8, 2)
    assert cable.dates.unlimited is None


def test_comic_digital_price(talker):
    cw1 = talker.comic(4216)
    assert cw1.title == "Civil War (2006) #1"
    assert cw1.prices.print == Decimal("0.00")
    assert cw1.prices.digital == Decimal("1.99")
    assert cw1.series.name == "Civil War (2006 - 2007)"
    assert cw1.format == "Comic"
    assert cw1.upc == "75960605921800111"
    assert cw1.issue_number == 1
    assert cw1.digital_id == 5486
    assert len(cw1.stories) == 2
    assert cw1.stories[0].id == 5872
    assert cw1.stories[0].type == "cover"
    assert cw1.stories[0].name == "1 of 7 - 7XLS"
    assert cw1.stories[1].id == 5873
    assert cw1.stories[1].type == "interiorStory"
    assert cw1.stories[1].name == "1 of 7 - 7XLS"
    assert cw1.dates.on_sale == date(2006, 5, 3)
    assert cw1.dates.foc is None
    assert cw1.dates.unlimited == date(2009, 8, 12)
    assert cw1.images[0] == "http://i.annihil.us/u/prod/marvel/i/mg/e/f0/511307b2f1200.jpg"
    assert cw1.images[1] == "http://i.annihil.us/u/prod/marvel/i/mg/6/f0/4f75b393338cf.jpg"


def test_comic_characters(talker):
    a1 = talker.comic_characters(67002)
    assert len(a1.character) == 8
    she_hulk = a1.character[6]
    assert she_hulk.id == 1009583
    assert she_hulk.name == "She-Hulk (Jennifer Walters)"
    assert she_hulk.resource_uri == "http://gateway.marvel.com/v1/public/characters/1009583"
    assert (
        she_hulk.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/7/20/527bb5d64599e.jpg"
    )
    assert len(she_hulk.comics) == 20
    assert len(she_hulk.events) == 15
    assert len(she_hulk.series) == 20
    assert len(she_hulk.stories) == 20


def test_comic_creators(talker):
    a1 = talker.comic_creators(67002)
    assert len(a1.creator) > 0
    jason = a1.creator[0]
    assert jason.id == 11463
    assert jason.first_name == "Jason"
    assert jason.last_name == "Aaron"
    assert jason.full_name == "Jason Aaron"
    assert jason.resource_uri == "http://gateway.marvel.com/v1/public/creators/11463"
    assert jason.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/7/10/5cd9c7870670e.jpg"
    assert len(jason.comics) == 20
    assert len(jason.events) == 10
    assert len(jason.series) == 20
    assert len(jason.stories) == 20


def test_comic_events(talker):
    sw1 = talker.comic_events(52447)
    assert len(sw1.events) == 1
    secret_wars = sw1.events[0]
    assert secret_wars.id == 323
    assert secret_wars.title == "Secret Wars (2015)"
    assert (
        secret_wars.thumbnail
        == "http://i.annihil.us/u/prod/marvel/i/mg/c/70/545be45c5d6cc.jpg"
    )
    assert secret_wars.resource_uri == "http://gateway.marvel.com/v1/public/events/323"
    assert secret_wars.start == date(2015, 5, 1)
    assert secret_wars.end == date(2015, 12, 31)
    assert len(secret_wars.characters) == 9
    assert len(secret_wars.comics) == 20
    assert len(secret_wars.creators) == 20
    assert len(secret_wars.series) == 20
    assert secret_wars.next.id == 332
    assert secret_wars.next.name == "Dead No More: The Clone Conspiracy"
    assert secret_wars.next.resource_uri == "http://gateway.marvel.com/v1/public/events/332"
    assert secret_wars.previous.id == 321
    assert secret_wars.previous.name == "Spider-Verse"
    assert (
        secret_wars.previous.resource_uri == "http://gateway.marvel.com/v1/public/events/321"
    )


def test_comic_stories(talker):
    aforce4 = talker.comic_stories(51206)
    assert len(aforce4.stories) == 2
    s = aforce4.stories[1]
    assert s.id == 113990
    assert s.type == "story"
    assert s.resource_uri == "http://gateway.marvel.com/v1/public/stories/113990"
    assert len(s.characters) == 0
    assert len(s.comics) == 1
    assert len(s.creators) == 3
    assert len(s.events) == 1
    assert s.original_issue.id == 51206
    assert s.original_issue.name == "A-Force (2015) #4"
    assert s.original_issue.resource_uri == "http://gateway.marvel.com/v1/public/comics/51206"
    assert len(s.series) == 1
    assert s.thumbnail is None



"""
Test Stories module.

This module contains tests for Stories objects.
"""
import pytest

from esak import exceptions


def test_known_story(talker):
    sm = talker.story(35505)
    assert sm.title == "Spider-Man!"
    assert sm.type == "story"
    assert sm.thumbnail is None

    assert sm.creators[0].name == "Steve Ditko"
    assert sm.creators[0].resource_uri == "http://gateway.marvel.com/v1/public/creators/32"
    assert sm.creators[0].role == "inker"
    assert sm.creators[1].name == "Stan Goldberg"
    assert sm.creators[1].role == "colorist"
    assert sm.creators[1].resource_uri == "http://gateway.marvel.com/v1/public/creators/962"

    assert len(sm.events) == 0

    assert sm.series[0].name == "Amazing Fantasy (1962)"
    assert sm.series[0].resource_uri == "http://gateway.marvel.com/v1/public/series/2987"
    assert sm.series[1].name == "AMAZING FANTASY OMNIBUS HC (2007)"
    assert sm.series[1].resource_uri == "http://gateway.marvel.com/v1/public/series/2707"

    assert sm.original_issue.id == 16926
    assert sm.original_issue.name == "Amazing Fantasy (1962) #15"
    assert sm.original_issue.resource_uri == "http://gateway.marvel.com/v1/public/comics/16926"

    assert sm.characters[0].id == 1009610
    assert sm.characters[0].name == "Spider-Man (Peter Parker)"
    assert (
        sm.characters[0].resource_uri
        == "http://gateway.marvel.com/v1/public/characters/1009610"
    )

    assert len(sm.comics) == 2
    assert sm.comics[0].id == 16926
    assert sm.comics[0].name == "Amazing Fantasy (1962) #15"
    assert sm.comics[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/16926"

    assert len(sm.characters) == 1
    assert sm.characters[0].id == 1009610
    assert sm.characters[0].name == "Spider-Man (Peter Parker)"
    assert sm.characters[0].role is None
    assert (
        sm.characters[0].resource_uri
        == "http://gateway.marvel.com/v1/public/characters/1009610"
    )


def test_bad_story(talker):
    with pytest.raises(exceptions.ApiError):
        talker.story(-1)


def test_stories_list(talker):
    stories_lst = talker.stories_list(
        {
            "orderBy": "modified",
        }
    )

    stories_iter = iter(stories_lst)
    assert (next(stories_iter).id) == 32039
    assert (next(stories_iter).id) == 41777
    assert (next(stories_iter).id) == 8186
    assert len(stories_lst) == 20



"""
Creator module.

This module provides the following classes:

- Creator
- CreatorSchema
"""
import itertools

from marshmallow import INCLUDE, Schema, fields, post_load, pre_load
from marshmallow.exceptions import ValidationError

from esak import exceptions, generic_summary


class Creator:
    """
    The Creator object contains information for creators.

    :param `**kwargs`: The keyword arguments is used for setting creator data from Marvel.
    """

    def __init__(self, **kwargs):
        """Intialize a new Creator."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CreatorsSchema(Schema):
    """Schema for the Creator API."""

    id = fields.Int()
    firstName = fields.Str(attribute="first_name")
    middleName = fields.Str(attribute="middle_name")
    lastName = fields.Str(attribute="last_name")
    suffix = fields.Str()
    fullName = fields.Str(attribute="full_name")
    modified = fields.DateTime()
    resourceURI = fields.Str(attribute="resource_uri")
    # urls
    thumbnail = fields.Url()
    series = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    stories = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    events = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    comics = fields.Nested(generic_summary.GenericSummarySchema, many=True)

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """
        Clean the data from Marvel.

        :param data: Data from Marvel response.

        :returns: Marvel Response
        :rtype: dict
        """
        if data.get("code", 200) != 200:
            raise exceptions.ApiError(data.get("status"))

        if "status" in data:
            data = data["data"]["results"][0]

        if "thumbnail" in data:
            data["thumbnail"] = f"{data['thumbnail']['path']}.{data['thumbnail']['extension']}"

        if "events" in data:
            data["events"] = data["events"]["items"]

        if "series" in data:
            data["series"] = data["series"]["items"]

        if "stories" in data:
            data["stories"] = data["stories"]["items"]

        if "comics" in data:
            data["comics"] = data["comics"]["items"]

        data["id"] = data["resourceURI"].split("/")[-1]
        return data

    @post_load
    def make(self, data, **kwargs):
        """
        Make the Creator object.

        :param data: Data from Marvel response.

        :returns: :class:`Creator` object
        :rtype: Creator
        """
        return Creator(**data)


class CreatorsList:
    """The CreatorsList object contains a list of `Creator` objects."""

    def __init__(self, response):
        """Initialize a new CreatorsList."""
        self.creator = []

        for series_dict in response["data"]["results"]:
            try:
                result = CreatorsSchema().load(series_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.creator.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.creator)

    def __len__(self):
        """Return the length of the object."""
        return len(self.creator)

    def __getitem__(self, index):
        """Return the object of a at index."""
        try:
            return next(itertools.islice(self.creator, index, index + 1))
        except TypeError:
            return list(itertools.islice(self.creator, index.start, index.stop, index.step))



"""Project entry file."""
__version__ = "1.2.0"

from typing import Optional

from esak import exceptions, session, sqlite_cache


def api(
    public_key: Optional[str] = None,
    private_key: Optional[str] = None,
    cache: Optional[sqlite_cache.SqliteCache] = None,
) -> session.Session:
    """Entry function the sets login credentials for Marvel's API.

    :param public_key: The username used for metron.cloud.
    :type public_key: str, optional
    :param private_key: The password used for metron.cloud.
    :type private_key: str, optional
    :param cache: SqliteCache to use
    :type cache: SqliteCache, optional

    :raises: :class:`AuthenticationError`
    """
    if public_key is None:
        raise exceptions.AuthenticationError("Missing public_key.")

    if private_key is None:
        raise exceptions.AuthenticationError("Missing private_key.")

    return session.Session(public_key, private_key, cache=cache)



"""
Test Series module.
This module contains tests for Series objects.
"""
import datetime
from decimal import Decimal

import pytest

from esak import comic, exceptions
from esak.series import Series


def test_known_series(talker):

    usms = talker.series(466)
    assert usms.title == "Ultimate Spider-Man (2000 - 2009)"
    assert usms.id == 466
    assert usms.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/6/c0/5149db8019dc9.jpg"
    assert usms.end_year == 2009
    assert usms.start_year == 2000
    assert usms.modified == datetime.datetime(
        2013,
        3,
        20,
        11,
        54,
        44,
        tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=72000)),
    )
    assert usms.description == (
        "In 2000, Marvel embarked on a bold new experiment, re-imagining some "
        "of their greatest heroes in the 21st century, beginning with Spider-Man! "
        "Writer Brian Michael Bendis along with artists Mark Bagley and Stuart Immonen "
        "invite you to discover the world of Peter Parker in a whole new way with the "
        "series that changed everything!"
    )
    assert usms.rating == "A"

    assert len(usms.comics) == 20
    assert usms.comics[0].id == 4372
    assert usms.comics[0].name == "Ultimate Spider-Man (2000) #1"
    assert usms.comics[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/4372"

    assert len(usms.characters) == 20
    assert usms.characters[11].id == 1010921
    assert usms.characters[11].name == "Doctor Octopus (Ultimate)"
    assert usms.characters[11].role is None
    assert (
        usms.characters[11].resource_uri
        == "http://gateway.marvel.com/v1/public/characters/1010921"
    )

    assert len(usms.creators) == 20
    assert usms.creators[12].id == 162
    assert usms.creators[12].name == "John Cassaday"
    assert usms.creators[12].role == "penciller"
    assert usms.creators[12].resource_uri == "http://gateway.marvel.com/v1/public/creators/162"


def test_bad_series(talker):
    with pytest.raises(exceptions.ApiError):
        talker.series(-1)


def test_bad_response_data():
    with pytest.raises(exceptions.ApiError):
        comic.ComicsList({"data": {"results": [{"modified": "potato"}]}})


def test_all_series(talker):
    # check known values
    usm = talker.series_list(
        params={"title": "Ultimate Spider-Man"}
    )  # don't include '(year - year)', it doesn't work
    assert usm is not None
    assert len(usm) == 1
    assert isinstance(usm[0], Series)
    # check the filter works returning multiple items
    series_iter = talker.series_list(
        params={"seriesType": "ongoing", "startYear": 2009, "limit": 10}
    )
    powers = series_iter[7]
    assert powers.id == 9073
    assert powers.start_year == 2009
    assert powers.end_year == 2012
    assert powers.title == "Powers (2009 - 2012)"
    assert powers.type == "ongoing"
    assert powers.resource_uri == "http://gateway.marvel.com/v1/public/series/9073"


def test_pulls_verbose(talker):
    series = talker.series_list()
    s_iter = iter(series)
    assert next(s_iter).id == 31445
    assert next(s_iter).id == 26024
    assert next(s_iter).id == 18454
    assert len(series) > 0


def test_series_characters(talker):
    sm = talker.series_characters(24396)
    assert len(sm.character) == 20
    kingpin = sm.character[10]
    assert kingpin.id == 1009389
    assert kingpin.name == "Kingpin"
    assert kingpin.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/9/60/526034fb5aff7.jpg"
    assert len(kingpin.comics) == 20
    assert len(kingpin.events) == 7
    assert len(kingpin.series) == 20
    assert len(kingpin.stories) == 20
    assert kingpin.urls.onsale_date is None


def test_series_comics(talker):
    sm = talker.series_comics(24396)
    assert len(sm.comics) == 20
    sm_75 = sm.comics[10]
    assert sm_75.format == "Comic"
    assert sm_75.title == "The Amazing Spider-Man (2018) #75 (Variant)"
    assert sm_75.issue_number == 75
    assert sm_75.page_count == 56
    assert sm_75.resource_uri == "http://gateway.marvel.com/v1/public/comics/97900"
    assert sm_75.upc == "75960608936907551"
    assert len(sm_75.creators) == 5
    assert len(sm_75.characters) == 0
    assert len(sm_75.events) == 0
    assert len(sm_75.stories) == 2
    assert sm_75.prices.print == Decimal("5.99")
    assert sm_75.prices.digital is None


def test_series_creators(talker):
    sm = talker.series_creators(24396)
    assert len(sm.creator) == 20
    bagley = sm.creator[10]
    assert bagley.id == 87
    assert bagley.first_name == "Mark"
    assert bagley.last_name == "Bagley"
    assert bagley.full_name == "Mark Bagley"
    assert bagley.resource_uri == "http://gateway.marvel.com/v1/public/creators/87"
    assert bagley.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/9/b0/4bc5d2f67f706.jpg"
    assert len(bagley.comics) == 20
    assert len(bagley.events) == 13
    assert len(bagley.series) == 20
    assert len(bagley.stories) == 20


def test_series_events(talker):
    AvX = talker.series_events(15305)
    assert len(AvX.events) == 1
    e = AvX[0]
    assert e.id == 310
    assert e.title == "Avengers VS X-Men"
    assert e.resource_uri == "http://gateway.marvel.com/v1/public/events/310"
    assert e.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/3/20/5109a1f93b543.jpg"
    assert len(e.characters) == 20
    assert len(e.comics) == 20
    assert len(e.creators) == 20
    assert len(e.series) == 9
    assert len(e.stories) == 20
    assert e.start == datetime.date(2012, 4, 4)
    assert e.end == datetime.date(2012, 9, 19)
    assert e.next.id == 311
    assert e.next.name == "Marvel NOW!"
    assert e.next.resource_uri == "http://gateway.marvel.com/v1/public/events/311"
    assert e.previous.id == 309
    assert e.previous.name == "Shattered Heroes"
    assert e.previous.resource_uri == "http://gateway.marvel.com/v1/public/events/309"


def test_series_stories(talker):
    AvX = talker.series_stories(15305)
    assert len(AvX.stories) == 20
    s = AvX.stories[13]
    assert s.id == 93246
    assert s.type == "story"
    assert s.resource_uri == "http://gateway.marvel.com/v1/public/stories/93246"
    assert len(s.characters) == 4
    assert len(s.comics) == 1
    assert len(s.creators) == 6
    assert len(s.events) == 1
    assert len(s.series) == 1
    assert s.original_issue.id == 41193
    assert s.original_issue.name == "Avengers Vs. X-Men (2012) #4"
    assert s.original_issue.resource_uri == "http://gateway.marvel.com/v1/public/comics/41193"



"""
Series module.

This module provides the following classes:

- Series
- SeriesSchema
"""
import itertools

from marshmallow import INCLUDE, Schema, fields, post_load, pre_load
from marshmallow.exceptions import ValidationError

from esak import exceptions, generic_summary


class Series:
    """
    The Series object contains information for a series.

    :param `**kwargs`: The keyword arguments is used for setting series data from Marvel.
    """

    def __init__(self, **kwargs):
        """Intialize a new series."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class SeriesSchema(Schema):
    """Schema for the Comic API."""

    id = fields.Int()
    title = fields.Str()
    description = fields.Str(allow_none=True)
    resourceURI = fields.Str(attribute="resource_uri")
    # urls
    startYear = fields.Int(attribute="start_year", allow_none=True)
    endYear = fields.Int(attribute="end_year", allow_none=True)
    rating = fields.Str(allow_none=True)
    modified = fields.DateTime()
    thumbnail = fields.URL(allow_none=True)
    comics = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    stories = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    events = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    characters = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    creators = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    next = fields.Nested(generic_summary.GenericSummarySchema, allow_none=True)
    previous = fields.Nested(generic_summary.GenericSummarySchema, allow_none=True)

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """
        Clean the data from Marvel.

        :param data: Data from Marvel response.

        :returns: Marvel Response
        :rtype: dict
        """
        if data.get("code", 200) != 200:
            raise exceptions.ApiError(data.get("status"))

        if "status" in data:
            data = data["data"]["results"][0]

        # Marvel series 6664, and maybe others, returns a modified of
        # "-0001-11-30T00:00:00-0500". The best way to handle this is
        # probably just to ignore it, since I don't know how to fix it.
        if data.get("modified", " ")[0] == "-":
            del data["modified"]

        # derive ID
        data["id"] = data["resourceURI"].split("/")[-1]

        if "thumbnail" in data and data["thumbnail"] is not None:
            data["thumbnail"] = f"{data['thumbnail']['path']}.{data['thumbnail']['extension']}"
        else:
            data["thumbnail"] = None

        if "comics" in data:
            data["comics"] = data["comics"]["items"]

        if "stories" in data:
            data["stories"] = data["stories"]["items"]

        if "events" in data:
            data["events"] = data["events"]["items"]

        if "characters" in data:
            data["characters"] = data["characters"]["items"]

        if "creators" in data:
            data["creators"] = data["creators"]["items"]

        return data

    @post_load
    def make(self, data, **kwargs):
        """
        Make the Series object.

        :param data: Data from Marvel response.

        :returns: :class:`Series` object
        :rtype: Seriess
        """
        return Series(**data)


class SeriesList:
    """The SeriesList object contains a list of `Series` objects."""

    def __init__(self, response):
        """Initialize a new SeriesList."""
        self.series = []

        for series_dict in response["data"]["results"]:
            try:
                result = SeriesSchema().load(series_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.series.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.series)

    def __len__(self):
        """Return the length of the object."""
        return len(self.series)

    def __getitem__(self, index):
        """Return the object of a at index."""
        try:
            return next(itertools.islice(self.series, index, index + 1))
        except TypeError:
            return list(itertools.islice(self.series, index.start, index.stop, index.step))



"""
Events module.

This module provides the following classes:

- Event
- EventSchema
- EventsList
"""
import itertools

from marshmallow import INCLUDE, Schema, fields, post_load, pre_load
from marshmallow.exceptions import ValidationError

from esak import exceptions, generic_summary


class Events:
    """
    The Event object contains information for events.

    :param `**kwargs`: The keyword arguments is used for setting event data from Marvel.
    """

    def __init__(self, **kwargs) -> None:
        """Intialize a new event."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class EventSchema(Schema):
    """Schema for the Event API."""

    id = fields.Int()
    title = fields.Str()
    description = fields.Str()
    resourceURI = fields.Str(attribute="resource_uri")
    # urls
    modified = fields.DateTime()
    start = fields.Date(allow_none=True)
    end = fields.Date(allow_none=True)
    thumbnail = fields.Url()
    comics = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    stories = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    series = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    characters = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    creators = fields.Nested(generic_summary.GenericSummarySchema, many=True)
    next = fields.Nested(generic_summary.GenericSummarySchema, allow_none=True)
    previous = fields.Nested(generic_summary.GenericSummarySchema, allow_none=True)

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        dateformat = "%Y-%m-%d %H:%M:%S"

    @pre_load
    def process_input(self, data, **kwargs):
        """
        Clean the data from Marvel.

        :param data: Data from Marvel response.

        :returns: Marvel Response
        :rtype: dict
        """
        if data.get("code", 200) != 200:
            raise exceptions.ApiError(data.get("status"))

        if "status" in data:
            data = data["data"]["results"][0]

        if "stories" in data:
            data["stories"] = data["stories"]["items"]

        if "characters" in data:
            data["characters"] = data["characters"]["items"]

        if "creators" in data:
            data["creators"] = data["creators"]["items"]

        if "comics" in data:
            data["comics"] = data["comics"]["items"]

        if "series" in data:
            data["series"] = data["series"]["items"]

        if "thumbnail" in data:
            data["thumbnail"] = f"{data['thumbnail']['path']}.{data['thumbnail']['extension']}"

        return data

    @post_load
    def make(self, data, **kargs):
        """
        Make the events object.

        :param data: Data from Marvel response.

        :returns: :class:`Events` object
        :rtype: Events
        """
        return Events(**data)


class EventsList:
    """The EventsList object contains a list of `Events` objects."""

    def __init__(self, response):
        """Initialize a new EventsList."""
        self.events = []

        for events_dict in response["data"]["results"]:
            try:
                result = EventSchema().load(events_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.events.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.events)

    def __len__(self):
        """Return the length of the object."""
        return len(self.events)

    def __getitem__(self, index):
        """Return the object of a at index."""
        try:
            return next(itertools.islice(self.events, index, index + 1))
        except TypeError:
            return list(itertools.islice(self.events, index.start, index.stop, index.step))



"""
Urls module.

This module provides the following classes:

- Urls
- UrlsSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load, pre_load


class Urls:
    """
    The Urls object contains commonly used urls.

    :param `digital_purchase_date`: The url for the digital_purchase_date
    :param `foc_date`: The url for the final order cutoff date.
    :param `onsale_date`: The url for the on sale date.
    :param `unlimited_date`: The url for the Marvel Unlimited date.
    :param `wiki`: The url for the wiki entry.
    :param `detail`: The url for detail information.
    :param `**kwargs`: The keyword arguments used for setting any other url from Marvel.
    """

    def __init__(
        self,
        digital_purchase_date=None,
        foc_date=None,
        onsale_date=None,
        unlimited_date=None,
        wiki=None,
        detail=None,
        **kwargs,
    ):
        """Intialize a new url."""
        self.digital_purchase_date = digital_purchase_date
        self.foc_date = foc_date
        self.onsale_date = onsale_date
        self.unlimited_date = unlimited_date
        self.wiki = wiki
        self.detail = detail
        self.unknown = kwargs


class UrlsSchema(Schema):
    """Schema for the Urls."""

    digitalPurchaseDate = fields.Url(attribute="digital_purchase_date")
    focDate = fields.Url(attribute="foc_date")
    onsaleDate = fields.Url(attribute="onsale_date")
    unlimitedDate = fields.Url(attribute="unlimited_date")
    # Should these go into a separate class like CharacterUrls?
    # For now let's put them here, but it may be something to consider to split them.
    wiki = fields.Url()
    detail = fields.Url()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """Clean up response data."""
        return {d["type"]: d["url"] for d in data}

    @post_load
    def make(self, data, **kwargs):
        """
        Make the urls object.

        :param data: Data from Marvel response.

        :returns: :class:`Urls` object
        :rtype: Urls
        """
        return Urls(**data)



"""
Prices module.

This module provides the following classes:

- Prices
- PricesSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load, pre_load


class Prices:
    """
    The Prices object contains price information.

    :param `print`: The price of print comic.
    :param `digital`: The price of digital comic.
    :param `**kwargs`: The keyword arguments used for setting any other price from Marvel.
    """

    def __init__(self, print=None, digital=None, **kwargs):
        """Intialize a new price."""
        self.print = print
        self.digital = digital
        self.unknown = kwargs


class PriceSchemas(Schema):
    """Schema for the Prices."""

    printPrice = fields.Decimal(places=2, attribute="print")
    digitalPurchasePrice = fields.Decimal(places=2, attribute="digital")

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """Clean up response data."""
        return {d["type"]: d["price"] for idx, d in enumerate(data, 0) if d["type"][idx]}

    @post_load
    def make(self, data, **kwargs):
        """
        Make the prices object.

        :param data: Data from Marvel response.

        :returns: :class:`Prices` object
        :rtype: Prices
        """
        return Prices(**data)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esak']

package_data = \
{'': ['*']}

install_requires = \
['marshmallow>=3.13.0,<4.0.0', 'requests>=2.26.0,<3.0.0']

extras_require = \
{'docs': ['sphinx-rtd-theme>=0.5.2,<0.6.0']}

setup_kwargs = {
    'name': 'esak',
    'version': '1.2.0',
    'description': 'Marvel API wrapper for python.',
    'long_description': 'esak - Marvel API wrapper for python 3\n===========================================\n\n\n.. image:: https://img.shields.io/pypi/v/esak.svg?logo=PyPI&label=Version&style=flat-square   :alt: PyPI\n    :target: https://pypi.org/project/esak\n\n.. image:: https://img.shields.io/pypi/pyversions/esak.svg?logo=Python&label=Python-Versions&style=flat-square\n    :target: https://pypi.org/project/esak\n\n.. image:: https://img.shields.io/github/license/bpepple/esak\n    :target: https://opensource.org/licenses/GPL-3.0\n\n.. image:: https://codecov.io/gh/bpepple/esak/branch/master/graph/badge.svg?token=L1EGNX24I2\n    :target: https://codecov.io/gh/bpepple/esak\n\n.. image:: https://img.shields.io/badge/Code%20Style-Black-000000.svg?style=flat-square\n    :target: https://github.com/psf/black\n\n- `Code on Github <https://github.com/bpepple/esak>`_\n- `Published on PyPi <https://pypi.python.org/pypi/esak>`_\n- `Marvel API documentation <https://developer.marvel.com/docs>`_\n\n**To install:**\n\n.. code-block:: bash\n\n    $ pip3 install --user esak\n \n**Example Usage:**\n\n.. code-block:: python\n\n    import esak\n\n    # Your own config file to keep your private key local and secret\n    from config import public_key, private_key\n\n    # Authenticate with Marvel, with keys I got from http://developer.marvel.com/\n    m = esak.api(public_key, private_key)\n\n    # Get all comics from this week, sorted alphabetically by title\n    pulls = sorted(m.comics_list({\n        \'format\': "comic",\n        \'formatType\': "comic",\n        \'noVariants\': True,\n        \'dateDescriptor\': "thisWeek",\n        \'limit\': 100}),\n        key=lambda comic: comic.title)\n\n    for comic in pulls:\n        # Write a line to the file with the name of the issue, and the\n        # id of the series\n        print(\'{} (series #{})\'.format(comic.title, comic.series.id))\n\nDocumentation\n-------------\n- `Read the project documentation <https://esak.readthedocs.io/en/stable/>`_\n\nBugs/Requests\n-------------\n  \nPlease use the `GitHub issue tracker <https://github.com/bpepple/esak/issues>`_ to submit bugs or request features.\n\nContributing\n------------\n\n- When running a new test for the first time, set the environment variables\n  ``PUBLIC_KEY`` and ``PRIVATE_KEY`` to any Marel API keys. The result will be\n  stored in the `tests/testing_mock.sqlite` database without your keys.\n\n',
    'author': 'Brian Pepple',
    'author_email': 'bdpepple@gmail.com',
    'maintainer': 'Brian Pepple',
    'maintainer_email': 'bdpepple@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Test Cache module.
This module contains tests for SqliteCache objects.
"""
import json

import pytest
import requests_mock

from esak import api, exceptions, sqlite_cache


class NoGet:
    def store(self, key, value):
        # This method should store key value pair
        return


class NoStore:
    def get(self, key):
        return None


def test_no_get(dummy_pubkey, dummy_privkey):
    m = api(public_key=dummy_pubkey, private_key=dummy_privkey, cache=NoGet())

    with pytest.raises(exceptions.CacheError):
        m.series(466)


def test_no_store(dummy_pubkey, dummy_privkey):
    m = api(public_key=dummy_pubkey, private_key=dummy_privkey, cache=NoStore())

    with requests_mock.Mocker() as r:
        r.get(
            "http://gateway.marvel.com:80/v1/public/series/466",
            text='{"response_code": 200}',
        )

        with pytest.raises(exceptions.CacheError):
            m.series(466)


def test_sql_store(dummy_pubkey, dummy_privkey):
    fresh_cache = sqlite_cache.SqliteCache(":memory:")
    test_cache = sqlite_cache.SqliteCache("tests/testing_mock.sqlite")

    m = api(public_key=dummy_pubkey, private_key=dummy_privkey, cache=fresh_cache)
    url = "http://gateway.marvel.com:80/v1/public/series/466"

    assert fresh_cache.get(url) is None

    try:
        with requests_mock.Mocker() as r:
            r.get(url, text=json.dumps(test_cache.get(url)))
            m.series(466)

        assert fresh_cache.get(url) is not None
    except TypeError:
        print(
            "This test will fail after cache db deleted.\n"
            "It should pass if you now re-run the test suite "
            "without deleting the database."
        )
        assert False



"""
Test Characters module.
This module contains tests for Character objects.
"""
from datetime import date

import pytest

from esak import exceptions


def test_known_character(talker):
    cap = talker.character(1009220)
    assert cap.name == "Captain America"
    assert cap.resource_uri == "http://gateway.marvel.com/v1/public/characters/1009220"
    assert cap.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/3/50/537ba56d31087.jpg"
    assert len(cap.series) > 0
    assert len(cap.events) > 0
    assert len(cap.stories) == 20
    assert cap.stories[0].id == 670
    assert cap.stories[0].name == "X-MEN (2004) #186"
    assert cap.stories[0].type == "cover"
    assert cap.stories[19].id == 1606
    assert cap.stories[19].name == "WEAPON X (2002) #14"
    assert cap.stories[19].type == "cover"

    assert len(cap.comics) == 20
    assert cap.comics[0].id == 43488
    assert cap.comics[0].name == "A+X (2012) #1"
    assert cap.comics[0].resource_uri == "http://gateway.marvel.com/v1/public/comics/43488"

    assert len(cap.series) == 20
    assert cap.series[10].id == 9865
    assert cap.series[10].name == "All-Winners Squad: Band of Heroes (2011)"
    assert cap.series[10].resource_uri == "http://gateway.marvel.com/v1/public/series/9865"


def test_bad_character(talker):
    with pytest.raises(exceptions.ApiError):
        talker.character(-1)


def test_pulls_verbose(talker):
    characters = talker.characters_list(
        {
            "orderBy": "modified",
        }
    )

    c_iter = iter(characters)
    assert (next(c_iter).name) == "Howard Saint"
    assert (next(c_iter).name) == "The Phantom"
    assert (next(c_iter).name) == "Nextwave"
    assert len(characters) > 0


def test_character_comics(talker):
    cap = talker.character_comics(1009220)
    assert len(cap.comics) == 20
    af2 = cap.comics[10]
    assert af2.id == 71246
    assert af2.title == "Amazing Fantasy (2021) #2"
    assert af2.issue_number == 2
    assert af2.page_count == 32
    assert af2.upc == "75960609104100211"
    assert af2.format == "Comic"
    assert af2.series.id == 25984
    assert af2.series.name == "Amazing Fantasy (2021 - Present)"
    assert af2.series.resource_uri == "http://gateway.marvel.com/v1/public/series/25984"
    assert af2.dates.foc == date(2021, 8, 2)
    assert af2.dates.on_sale == date(2021, 8, 25)
    assert len(af2.stories) == 2
    assert len(af2.events) == 0
    assert len(af2.creators) == 3


def test_character_events(talker):
    cap = talker.character_events(1009220)
    assert len(cap.events) == 20
    fall = cap.events[9]
    assert fall.id == 248
    assert fall.title == "Fall of the Mutants"
    assert fall.thumbnail == "http://i.annihil.us/u/prod/marvel/i/mg/8/a0/51cb2f521ae35.jpg"
    assert fall.start == date(1988, 1, 10)
    assert fall.end == date(2007, 1, 17)
    assert fall.next.id == 252
    assert fall.next.name == "Inferno"
    assert fall.previous.id == 246
    assert fall.previous.name == "Evolutionary War"


def test_character_series(talker):
    cap = talker.character_series(1009220)
    assert len(cap.series) == 20
    af = cap.series[12]
    assert af.title == "Amazing Fantasy (2021 - Present)"
    assert af.id == 25984
    assert af.start_year == 2021
    assert af.end_year == 2099
    assert len(af.comics) == 14
    assert len(af.characters) == 3
    assert len(af.creators) == 7
    assert af.next is None
    assert af.previous is None


def test_character_stories(talker):
    cap = talker.character_stories(1009220)
    assert len(cap.stories) == 20
    av_503 = cap.stories[5]
    assert av_503.id == 1042
    assert av_503.title == "Avengers (1998) #503"
    assert av_503.type == "cover"
    assert len(av_503.characters) == 1
    assert len(av_503.comics) == 1
    assert len(av_503.creators) == 1
    assert av_503.original_issue.id == 923
    assert av_503.original_issue.name == "Avengers (1998) #503"
    assert (
        av_503.original_issue.resource_uri == "http://gateway.marvel.com/v1/public/comics/923"
    )



"""
Session module.

This module provides the following classes:

- Session
"""
import datetime
import hashlib
import platform
import urllib.parse
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union

import requests
from marshmallow import ValidationError

# Alias these modules to prevent namespace collision with methods.
from esak import __version__
from esak import character as ch
from esak import comic as com
from esak import creator as cr
from esak import event as ev
from esak import exceptions
from esak import series as ser
from esak import sqlite_cache, stories


class Session:
    """
    Session to request api endpoints.

    :param str public_key: The public_key for authentication with Marvel
    :param str private_key: The private_key used for authentication with Marvel
    :param SqliteCache optional: SqliteCache to use
    """

    def __init__(
        self,
        public_key: str,
        private_key: str,
        cache: Optional[sqlite_cache.SqliteCache] = None,
    ):
        """Intialize a new Session."""
        self.public_key = public_key
        self.private_key = private_key
        self.cache = cache
        self.api_url = "http://gateway.marvel.com:80/v1/public/{}"

    def _create_cached_params(self, params: Dict[str, Any]) -> str:
        # Generate part of cache key before hash, apikey and timestamp added
        cache_params = ""
        if params:
            orderedParams = OrderedDict(sorted(params.items(), key=lambda t: t[0]))
            cache_params = f"?{urllib.parse.urlencode(orderedParams)}"
        return cache_params

    def _create_auth_hash(self, now_string: str) -> str:
        auth_hash = hashlib.md5()
        auth_hash.update(now_string.encode("utf-8"))
        auth_hash.update(self.private_key.encode("utf-8"))
        auth_hash.update(self.public_key.encode("utf-8"))
        return auth_hash.hexdigest()

    def _update_params(self, params: Dict[str, Any]) -> None:
        now_string = datetime.datetime.now().strftime("%Y-%m-%d%H:%M:%S")

        params["hash"] = self._create_auth_hash(now_string)
        params["apikey"] = self.public_key
        params["ts"] = now_string

    def _get_results_from_cache(self, key: str) -> Optional[Any]:
        cached_response = None

        if self.cache:
            try:
                cached_response = self.cache.get(key)
                if cached_response is not None:
                    return cached_response
            except AttributeError as e:
                raise exceptions.CacheError(
                    f"Cache object passed in is missing attribute: {repr(e)}"
                )

        return cached_response

    def _save_results_to_cache(self, key: str, data: str) -> None:
        if self.cache:
            try:
                self.cache.store(key, data)
            except AttributeError as e:
                raise exceptions.CacheError(
                    f"Cache object passed in is missing attribute: {repr(e)}"
                )

    def _call(self, endpoint: List[Union[str, int]], params: Dict[str, Any] = None) -> Any:
        if params is None:
            params = {}

        url = self.api_url.format("/".join(str(e) for e in endpoint))
        cache_params = self._create_cached_params(params)
        cache_key = f"{url}{cache_params}"
        cached_response = self._get_results_from_cache(cache_key)

        if cached_response is not None:
            return cached_response

        header = {
            "User-Agent": f"esak/{__version__} ({platform.system()}; {platform.release()})"
        }
        self._update_params(params)
        response = requests.get(
            url,
            params=params,
            headers=header,
        )

        data = response.json()

        if "message" in data:
            raise exceptions.ApiError(data["message"])

        if response.status_code == 200:
            self._save_results_to_cache(cache_key, data)

        return data

    def comic(self, _id: int) -> com.Comic:
        """
        Request data for a comic based on it's ``_id``.

        :param int _id: The comic id.

        :return: :class:`Comic` object
        :rtype: Comic

        :raises: :class:`ApiError`
        """
        try:
            return com.ComicSchema().load(self._call(["comics", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def comic_characters(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ch.CharactersList:
        """
        Request a list of characters from a comic.

        :param int _id: The comic id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Character` objects.
        :rtype: CharactersList
        """
        if params is None:
            params = {}

        return ch.CharactersList(self._call(["comics", _id, "characters"], params=params))

    def comic_creators(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> cr.CreatorsList:
        """
        Request a list of creators from a comic.

        :param int _id: The comic id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Creator` objects.
        :rtype: CreatorsList
        """
        if params is None:
            params = {}

        return cr.CreatorsList(self._call(["comics", _id, "creators"], params=params))

    def comic_events(self, _id: int, params: Optional[Dict[str, Any]] = None) -> ev.EventsList:
        """
        Request a list of events from a comic.

        :param int _id: The comic id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Event` objects.
        :rtype: EventsList
        """
        if params is None:
            params = {}

        return ev.EventsList(self._call(["comics", _id, "events"], params=params))

    def comic_stories(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> stories.StoriesList:
        """
        Request a list of stories from a comic.

        :param int _id: The comic id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Stories` objects.
        :rtype: StoriesList
        """
        if params is None:
            params = {}

        return stories.StoriesList(self._call(["comics", _id, "stories"], params=params))

    def comics_list(self, params: Optional[Dict[str, Any]] = None) -> com.ComicsList:
        """
        Request a list of comics.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        if params is None:
            params = {}

        return com.ComicsList(self._call(["comics"], params=params))

    def series(self, _id: int) -> ser.Series:
        """
        Request data for a series based on it's ``_id``.

        :param int _id: The series id.

        :return: :class:`Series` object
        :rtype: Series

        :raises: :class:`ApiError`
        """
        try:
            return ser.SeriesSchema().load(self._call(["series", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def series_characters(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ch.CharactersList:
        """
        Request a list of characters from a series.

        :param int _id: The series id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Character` objects.
        :rtype: CharactersList
        """
        if params is None:
            params = {}

        return ch.CharactersList(self._call(["series", _id, "characters"], params=params))

    def series_comics(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> com.ComicsList:
        """
        Request a list of comics from a series.

        :param int _id: The series id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        if params is None:
            params = {}

        return com.ComicsList(self._call(["series", _id, "comics"], params=params))

    def series_creators(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> cr.CreatorsList:
        """
        Request a list of creators from a series.

        :param int _id: The series id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Creator` objects.
        :rtype: CreatorsList
        """
        if params is None:
            params = {}

        return cr.CreatorsList(self._call(["series", _id, "creators"], params=params))

    def series_events(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ev.EventsList:
        """
        Request a list of events from a series.

        :param int _id: The series id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Event` objects.
        :rtype: EventsList
        """
        if params is None:
            params = {}

        return ev.EventsList(self._call(["series", _id, "events"], params=params))

    def series_stories(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> stories.StoriesList:
        """
        Request a list of stories from a series.

        :param int _id: The series id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Stories` objects.
        :rtype: StoriesList
        """
        if params is None:
            params = {}

        return stories.StoriesList(self._call(["series", _id, "stories"], params=params))

    def series_list(self, params: Optional[Dict[str, Any]] = None) -> ser.SeriesList:
        """
        Request a list of series.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Series` objects.
        :rtype: SeriesList
        """
        if params is None:
            params = {}

        return ser.SeriesList(self._call(["series"], params=params))

    def creator(self, _id: int) -> cr.Creator:
        """
        Request data for a creator based on it's ``_id``.

        :param int _id: The creator id.

        :return: :class:`Creator` object
        :rtype: Creator

        :raises: :class:`ApiError`
        """
        try:
            return cr.CreatorsSchema().load(self._call(["creators", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def creator_comics(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> com.ComicsList:
        """
        Request a list of comics from a creator.

        :param int _id: The creator id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        if params is None:
            params = {}

        return com.ComicsList(self._call(["creators", _id, "comics"], params=params))

    def creator_events(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ev.EventsList:
        """
        Request a list of events from a creator.

        :param int _id: The creator id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Event` objects.
        :rtype: EventsList
        """
        if params is None:
            params = {}

        return ev.EventsList(self._call(["creators", _id, "events"], params=params))

    def creator_series(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ser.SeriesList:
        """
        Request a list of series by a creator.

        :param int _id: The creator id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Series` objects.
        :rtype: SeriesList
        """
        if params is None:
            params = {}

        return ser.SeriesList(self._call(["creators", _id, "series"], params=params))

    def creator_stories(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> stories.StoriesList:
        """
        Request a list of stories from a creator.

        :param int _id: The creator id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Stories` objects.
        :rtype: StoriesList
        """
        if params is None:
            params = {}

        return stories.StoriesList(self._call(["creators", _id, "stories"], params=params))

    def creators_list(self, params: Optional[Dict[str, Any]] = None) -> cr.CreatorsList:
        """
        Request a list of creators.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Creator` objects.
        :rtype: CreatorsList
        """
        if params is None:
            params = {}

        return cr.CreatorsList(self._call(["creators"], params=params))

    def character(self, _id: int) -> ch.Character:
        """
        Request data for a character based on it's ``_id``.

        :param int _id: The character id.

        :return: :class:`Character` object
        :rtype: Character

        :raises: :class:`ApiError`
        """
        try:
            return ch.CharacterSchema().load(self._call(["characters", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def character_comics(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> com.ComicsList:
        """
        Request a list of comics for a character.

        :param int _id: The character id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        if params is None:
            params = {}

        return com.ComicsList(self._call(["characters", _id, "comics"], params=params))

    def character_events(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ev.EventsList:
        """
        Request a list of events for a character.

        :param int _id: The character id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Event` objects.
        :rtype: EventsList
        """
        if params is None:
            params = {}

        return ev.EventsList(self._call(["characters", _id, "events"], params=params))

    def character_series(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> ser.SeriesList:
        """
        Request a list of series for a character.

        :param int _id: The character id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Series` objects.
        :rtype: SeriesList
        """
        if params is None:
            params = {}

        return ser.SeriesList(self._call(["characters", _id, "series"], params=params))

    def character_stories(
        self, _id: int, params: Optional[Dict[str, Any]] = None
    ) -> stories.StoriesList:
        """
        Request a list of stories for a character.

        :param int _id: The character id.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Stories` objects.
        :rtype: StoriesList
        """
        if params is None:
            params = {}

        return stories.StoriesList(self._call(["characters", _id, "stories"], params=params))

    def characters_list(self, params: Optional[Dict[str, Any]] = None) -> ch.CharactersList:
        """
        Request a list of characters.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Character` objects.
        :rtype: CharactersList
        """
        if params is None:
            params = {}

        return ch.CharactersList(self._call(["characters"], params=params))

    def story(self, _id: int) -> stories.Stories:
        """
        Request data for a Story based on it's ``_id``.

        :param int _id: The story id.

        :return: :class:`Stories` object
        :rtype: Stories

        :raises: :class:`ApiError`
        """
        try:
            return stories.StoriesSchema().load(self._call(["stories", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def stories_list(self, params: Optional[Dict[str, Any]] = None) -> stories.StoriesList:
        """
        Request a list of stories.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Stories` objects.
        :rtype: StoriesList
        """
        if params is None:
            params = {}

        return stories.StoriesList(self._call(["stories"], params=params))

    def event(self, _id: int) -> ev.Events:
        """
        Request data for an event based on it's ``_id``.

        :param int _id: The event id.

        :return: :class:`Event` object
        :rtype: Event

        :raises: :class:`ApiError`
        """
        try:
            return ev.EventSchema().load(self._call(["events", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

    def events_list(self, params: Optional[Dict[str, Any]] = None) -> ev.EventsList:
        """
        Request a list of events.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Event` objects.
        :rtype: EventsList
        """
        if params is None:
            params = {}

        return ev.EventsList(self._call(["events"], params=params))
